import { Instance } from './Instance'
import { NetworkEvent } from '../common/binary/NetworkEvent'
import { User, UserConnectionState } from './User'
import { BinarySection } from '../common/binary/BinarySection'
import { EngineMessage } from '../common/EngineMessage'
import readEngineMessage from '../binary/message/readEngineMessage'
import readMessage from '../binary/message/readMessage'

export interface INetworkEvent {
    type: NetworkEvent
    user: User
    commands?: any
    clientTick?: number
}

export class InstanceNetwork {
    instance: Instance

    constructor(instance: Instance) {
        this.instance = instance
    }

    onRequest() {
        // TODO
    }

    onOpen(user: User) {
        user.connectionState = UserConnectionState.OpenPreHandshake
        user.network = this
    }

    async onHandshake(user: User, handshake: any) {
        try {
            user.connectionState = UserConnectionState.OpenAwaitingHandshake
            const connectionAccepted = await this.instance.onConnect(handshake)

            // @ts-ignore typescript is wrong that connectionState does not change, it changes during the await
            if (user.connectionState === UserConnectionState.Closed) {
                throw new Error('Connection closed before handshake completed.')
            }

            user.connectionState = UserConnectionState.Open

            // allow
            const bw = user.networkAdapter.createBufferWriter(3)
            bw.writeUInt8(BinarySection.EngineMessages)
            bw.writeUInt8(1)
            bw.writeUInt8(EngineMessage.ConnectionAccepted)

            user.send(bw.buffer)
            user.instance = this.instance
            this.onConnectionAccepted(user, connectionAccepted)
        } catch (err: any) {
            this.onConnectionDenied(user, err)

            // NOTE: we are keeping the code between these cases duplicated
            // if these do turn out to be identical in production we will clean it up
            // but for now I am suspicious that there will be different logic
            // in each of these later

            if (user.connectionState === UserConnectionState.OpenAwaitingHandshake) {
                // developer's code decided to reject this connection (rejected promise)
                const jsonErr = JSON.stringify(err)
                const denyReasonByteLength = Buffer.byteLength(jsonErr, 'utf8')

                // deny and send reason
                const bw = user.networkAdapter.createBufferWriter(3 + 4 /* string length 32 bits */ + denyReasonByteLength /* length of actual string*/)
                //binaryWriterCtor.create(3 + 4 /* string length 32 bits */ + denyReasonByteLength /* length of actual string*/)
                bw.writeUInt8(BinarySection.EngineMessages)
                bw.writeUInt8(1)
                bw.writeUInt8(EngineMessage.ConnectionDenied)
                bw.writeString(jsonErr)
                user.send(bw.buffer)
            }

            if (user.connectionState === UserConnectionState.Open) {
                // a loss of connection after handshake is complete
                const jsonErr = JSON.stringify(err)
                const denyReasonByteLength = Buffer.byteLength(jsonErr, 'utf8')

                // deny and send reason
                const bw = user.networkAdapter.createBufferWriter(3 + 4 /* string length 32 bits */ + denyReasonByteLength /* length of actual string*/)
                bw.writeUInt8(BinarySection.EngineMessages)
                bw.writeUInt8(1)
                bw.writeUInt8(EngineMessage.ConnectionDenied)
                bw.writeString(jsonErr)
                user.send(bw.buffer)
            }
        }
    }

    onMessage(user: User, buffer: Buffer | ArrayBuffer) {

        try {
            const binaryReader = user.networkAdapter.createBufferReader(buffer)
            const commands: any[] = []

            const commandSet = {
                type: NetworkEvent.CommandSet,
                user,
                commands,
                clientTick: -1
            }

            while (binaryReader.offset < binaryReader.byteLength) {
                const section = binaryReader.readUInt8()

                switch (section) {
                case BinarySection.EngineMessages: {
                    const count = binaryReader.readUInt8()
                    for (let i = 0; i < count; i++) {
                        const msg: any = readEngineMessage(binaryReader, this.instance.context)

                        if (msg.ntype === EngineMessage.ConnectionAttempt) {
                            const handshake = JSON.parse(msg.handshake)
                            this.onHandshake(user, handshake)
                        }

                        if (msg.ntype === EngineMessage.ClientTick) {
                            const clientTick = msg.tick
                            user.lastReceivedClientTick = clientTick
                            commandSet.clientTick = clientTick
                        }

                        if (msg.ntype === EngineMessage.Pong) {
                            user.calculateLatency()
                        }
                    }

                    break
                }
                case BinarySection.Commands: {
                    const count = binaryReader.readUInt8()
                    for (let i = 0; i < count; i++) {
                        const msg = readMessage(binaryReader, this.instance.context)
                        commands.push(msg)
                    }
                    break
                }
                case BinarySection.Requests: {
                    const count = binaryReader.readUInt8()
                    for (let i = 0; i < count; i++) {
                        const requestId = binaryReader.readUInt32()
                        const endpoint = binaryReader.readUInt32()
                        const str = binaryReader.readString()
                        const body = JSON.parse(str)
                        const cb = this.instance.responseEndPoints.get(endpoint)
                        if (cb) {
                            cb({ user, body }, (response: any) => {
                                user.responseQueue.push({
                                    requestId,
                                    response: JSON.stringify(response)
                                })
                            })
                        }
                    }
                    break
                }
                default: {
                    console.log('network hit default case while reading')
                    break
                }
                }
            }

            this.instance.queue.enqueue(commandSet)
        } catch (err) {
            // TODO there should be a way for a user to capture this error, perhaps a handler
            //console.log('on message err triggered', err)
            try {
                user.networkAdapter.disconnect(user, {})
            } catch (err2) {
                // TODO this is only in the case of an error while disconnecting
                // can these really occur?
            }
        }

    }

    onConnectionAccepted(user: User, payload: any) {
        user.network = this
        user.id = ++this.instance.incrementalUserId
        this.instance.users.set(user.id, user)

        this.instance.queue.enqueue({
            type: NetworkEvent.UserConnected,
            user,
            payload
        })
    }

    onConnectionDenied(user: User, payload: any) {
        this.instance.queue.enqueue({
            type: NetworkEvent.UserConnectionDenied,
            user,
            payload
        })
    }

    onClose(user: User) {
        if (user.connectionState === UserConnectionState.Open) {
            this.instance.queue.enqueue({
                type: NetworkEvent.UserDisconnected,
                user,
            })
            this.instance.users.delete(user.id)
        }
        user.connectionState = UserConnectionState.Closed
    }
}